# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from ..resource import options


def _method_to_operation(method):
    if method == 'GET':
        return set(['read'])

    if method == 'PUT':
        return set(['update', 'create', 'delete'])

    if method == 'POST':
        return set(['create'])

    if method == 'PATCH':
        return set(['update', 'create'])

    if method == 'DELETE':
        return set(['destroy'])


def _methods_to_operations(methods):
    operations = set()
    for method in methods:
        operations.update(_method_to_operation(method))

    return operations


def _operation_to_method(operation):
    if operation == 'read':
        return set(['GET'])

    if operation == 'update':
        return set(['PUT', 'PATCH', 'LINK', 'UNLINK'])

    if operation == 'create':
        return set(['PUT', 'PATCH', 'POST'])

    if operation == 'destroy':
        return set(['PUT', 'DELETE'])


def _operations_to_methods(operations):
    methods = set(['HEAD', 'OPTIONS'])
    for operation in operations:
        methods.update(_operation_to_method(operation))

    return methods


class ManagedResourceOptions(options.ResourceOptions):

    def __init__(self, meta, name, data, bases):
        # Initalize base resource options.
        super(ManagedResourceOptions, self).__init__(meta, name, data, bases)

        #! List of allowed operations.
        #! Resource operations are meant to generalize and blur the
        #! differences between "PATCH and PUT", "PUT = create / update",
        #! etc.
        #!
        #! If not provided and http_allowed_methods was provided instead
        #! the methods are appropriately mapped; else, the default
        #! configuration is provided.
        self.allowed_operations = meta.get('allowed_operations')
        if self.allowed_operations is None:
            if meta.get('http_allowed_methods'):
                self.allowed_operations = _methods_to_operations(meta.get(
                    'http_allowed_methods'))

            else:
                self.allowed_operations = (
                    'read',
                    'create',
                    'update',
                    'destroy',
                )

        # Coerce http allowed methods from the
        # allowed operations.
        if meta.get('http_allowed_methods') is None:
            if meta.get('allowed_operations'):
                self.http_allowed_methods = _operations_to_methods(meta.get(
                    'allowed_operations'))

        #! List of allowed HTTP methods against a whole
        #! resource (eg /user); if undeclared or None, will be defaulted
        #! to `http_allowed_methods`.
        self.http_list_allowed_methods = meta.get(
            'http_list_allowed_methods')

        if self.http_list_allowed_methods is None:
            if meta.get('list_allowed_operations'):
                self.http_list_allowed_methods = _operations_to_methods(
                    meta.get('list_allowed_operations'))

            else:
                self.http_list_allowed_methods = self.http_allowed_methods

        #! List of allowed HTTP methods against a single
        #! resource (eg /user/1); if undeclared or None, will be defaulted
        #! to `http_allowed_methods`.
        self.http_detail_allowed_methods = meta.get(
            'http_detail_allowed_methods')

        if self.http_detail_allowed_methods is None:
            if meta.get('detail_allowed_operations'):
                self.http_detail_allowed_methods = _operations_to_methods(
                    meta.get('detail_allowed_operations'))

            else:
                self.http_detail_allowed_methods = self.http_allowed_methods

        #! List of allowed operations against a whole resource.
        #! If undeclared or None, will be defaulted to `allowed_operations`.
        self.list_allowed_operations = meta.get('list_allowed_operations')

        if self.list_allowed_operations is None:
            if meta.get('http_list_allowed_methods'):
                self.list_allowed_operations = _methods_to_operations(
                    meta.get('http_list_allowed_methods'))

            else:
                self.list_allowed_operations = self.allowed_operations

        #! List of allowed operations against a single resource.
        #! If undeclared or None, will be defaulted to `allowed_operations`.
        self.detail_allowed_operations = meta.get(
            'detail_allowed_operations')

        if self.detail_allowed_operations is None:
            if meta.get('http_detail_allowed_methods'):
                self.detail_allowed_operations = _methods_to_operations(
                    meta.get('http_detail_allowed_methods'))

            else:
                self.detail_allowed_operations = self.allowed_operations

        #! Attribute to use for the slug or url segment
        #! that identifies the resource. The slug attribute is
        #! a special attribute; there are a couple of requirements.
        #! One is that it must be a unique reference. A /url/slug must
        #! return at most one item. Second is that as it is a special
        #! attribute that is not part of the body there is not
        #! a `prepare_slug` method.
        self.slug = meta.get('slug')
        if self.slug is None:
            # The slug defaults to `id`; which on most model engines
            # is the primary key. This is as good as a default as any I
            # suppose.
            self.slug = 'id'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import sys
from importlib import import_module
from wsgi_intercept.httplib2_intercept import install
import pytest
import armet
from armet import test


class BaseResourceTest(object):

    #! Host at which the intercept hook is installed.
    host = 'localhost'

    #! Port at which the intercept hook is installed.
    port = 5000

    @classmethod
    def setup_class(cls):
        # Initialize the test client.
        cls.client = test.Client(cls.host, cls.port)

    @pytest.fixture(autouse=True, scope='class')
    def initialize(self, request, connectors):
        # Install the WSGI interception layer.
        install()

        # Unload django.
        for module in list(sys.modules.keys()):
            if module and 'django' in module:
                del sys.modules[module]

        # Reset and clear all global cache in armet.
        from armet import decorators

        decorators._resources = {}
        # decorators._handlers = {}
        armet.use.config = {}

        # Re-initialize the configuration.
        armet.use(connectors=connectors, debug=True)

        prefix = 'tests.connectors.'
        callback = None
        if 'model' in connectors:
            # Initialize the database access layer.
            model = import_module(prefix + connectors['model'])
            callback = model.model_setup

            # Add the models module so that it can be generically imported.
            sys.modules[prefix + 'models'] = model
            request.cls.models = model

        # Initialize the http access layer.
        http = import_module(prefix + connectors['http'])
        http.http_setup(connectors, self.host, self.port, callback=callback)

        # Add a finalizer to teardown the http layer.
        request.addfinalizer(lambda: http.http_teardown(self.host, self.port))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import re
import six
from importlib import import_module
from armet import utils, authentication, authorization
from armet.exceptions import ImproperlyConfigured
from armet import connectors as included_connectors


def _merge(options, name, bases, default=None):
    """Merges a named option collection."""
    result = None
    for base in bases:
        if base is None:
            continue

        value = getattr(base, name, None)
        if value is None:
            continue

        result = utils.cons(result, value)

    value = options.get(name)
    if value is not None:
        result = utils.cons(result, value)

    return result or default


class ResourceOptions(object):

    def __init__(self, meta, name, data, bases):
        """
        Initializes the options object and defaults configuration not
        specified.

        @param[in] meta
            Dictionary of the merged meta attributes.

        @param[in] name
            Name of the resource class this is being instantiataed for.
        """
        #! Whether to allow display of debugging information
        #! to the client.
        self.debug = meta.get('debug')
        if self.debug is None:
            self.debug = False

        #! Whether to not actualize a resource from the described class.
        #! Abstract resources are meant as generic base classes.
        #!
        #! @note
        #!      Abstract only counts if it is directly set on the resource.
        self.abstract = data.get('abstract')

        #! Name of the resource to use in URIs; defaults to the dasherized
        #! version of the camel cased class name (eg. SomethingHere becomes
        #! something-here). The defaulted version also strips a trailing
        #! Resource from the name (eg. SomethingHereResource still becomes
        #! something-here).
        self.name = meta.get('name')
        if self.name is None:
            # Generate a name for the resource if one is not provided.
            # PascalCaseThis => pascal-case-this
            dashed = utils.dasherize(name).strip()
            if dashed:
                # Strip off a trailing Resource as well.
                self.name = re.sub(r'-resource$', '', dashed)

            else:
                # Something went wrong; just use the class name
                self.name = name

        elif callable(self.name):
            # If name is callable; pass in the class name and use what
            # we got back.
            self.name = self.name(name)

        #! True if the resource is expected to operate asynchronously.
        #!
        #! The side-effect of setting this to True is that returning from
        #! `dispatch` (and by extension `get`, `post`, etc.) does not
        #! terminate the connection. You must invoke `response.close()` to
        #! terminate the connection.
        self.asynchronous = meta.get('asynchronous', False)

        #! Connectors to use to connect to the environment.
        #!
        #! This is a dictionary that maps hooks (keys) to the connector to use
        #! for the hook.
        #!
        #! There is only 1 hook available currently, 'http', and it is
        #! required.
        #!
        #! The available connectors are as follows:
        #!  - http:
        #!      > django
        #!      > flask
        #!
        #! They may be used as follows:
        #!
        #! @code
        #! from armet import resources
        #! class Resource(resources.Resource):
        #!     class Meta:
        #!         connectors = {'http': 'django'}
        #! @endcode
        #!
        #! Connectors may also be specified as full paths to the connector
        #! module (if a connector is located somewhere other than
        #! armet.connectors) as follows:
        #!
        #! @code
        #! from armet import resources
        #! class Resource(resources.Resource):
        #!     class Meta:
        #!         connectors = {'http': 'some.other.place'}
        #! @endcode
        self.connectors = connectors = _merge(meta, 'connectors', bases, {})

        if not connectors.get('http') and not self.abstract:
            raise ImproperlyConfigured('No valid HTTP connector was detected.')

        # Pull out the connectors and convert them into module references.
        for key in connectors:
            connector = connectors[key]
            if isinstance(connector, six.string_types):
                if connector in getattr(included_connectors, key):
                    # Included shortname; prepend base.
                    connectors[key] = 'armet.connectors.{}'.format(connector)

        #! Additional options to handle and merge into the meta object
        #! at the class object level.
        #!
        #! This should be a simple set/list/tuple of options.
        #!
        #! @code
        #! from armet import resources
        #! class Resource(resources.Resource):
        #!     class Meta:
        #!         options = {'color', 'plant'}
        #!
        #! @endcode
        self.options = options = _merge(meta, 'options', bases, {})
        for name in options:
            # Pull out each option and stick it on the meta.
            setattr(self, name, meta.get(name))

        #! Regular-expression patterns to apply to the request path
        #! and pull arguments and traversal out of it.
        #!
        #! @code
        #! from armet import resources
        #! class Resource(resources.Resource):
        #!     class Meta:
        #!         patterns = [
        #!             # Match nothing after the resource name.
        #!             r'^$',
        #!
        #!             # Match a slug after the resource name.
        #!             r'^/(?P<slug>[^/]+)/?$',
        #!         ]
        #!
        #! @endcode
        #!
        #! Named parameters get auto-attached to `self` on the resource
        #! instance.
        #!
        #! This may be a list of lists/tuples as well to indicate simple
        #! traversal.
        #!
        #! @code
        #! from armet import resources
        #! class Resource(resources.Resource):
        #!     class Meta:
        #!         patterns = [
        #!             # Match nothing and don't traverse.
        #!             (None, r'^$'),
        #!
        #!             # Match the word 'user' and traverse.
        #!             # The remainder of the path is taken or
        #!             # the named gruop, "path" (if the whole path was
        #!             # matched) or the last-matched group (if none are
        #!             # named "path")
        #!             (UserResource, r'^/user'),
        #!         ]
        #!
        #! @endcode
        self.patterns = meta.get('patterns', [])
        for index, pattern in enumerate(self.patterns):
            # Coerce simple form.
            if isinstance(pattern, six.string_types):
                pattern = (None, pattern)

            # Compile the expression.
            self.patterns[index] = (pattern[0], re.compile(pattern[1]))

        #! Trailing slash handling.
        #! The value indicates which URI is the canonical URI and the
        #! alternative URI is then made to redirect (with a 301) to the
        #! canonical URI.
        self.trailing_slash = meta.get('trailing_slash', True)

        #! List of allowed HTTP methods.
        #! If not provided and allowed_operations was provided instead
        #! the operations are appropriately mapped; else, the default
        #! configuration is provided.
        self.http_allowed_methods = meta.get('http_allowed_methods')
        if self.http_allowed_methods is None:
            self.http_allowed_methods = (
                'HEAD',
                'OPTIONS',
                'GET',
                'POST',
                'PUT',
                'PATCH',
                'DELETE',
                'LINK',
                'UNLINK'
            )

        #! List of allowed HTTP headers.
        #! This is used only to request or prevent CORS requests.
        self.http_allowed_headers = meta.get('http_allowed_headers')
        if self.http_allowed_headers is None:
            self.http_allowed_headers = (
                'Content-Type',
                'Authorization',
                'Accept',
                'Origin'
            )

        #! List of exposed HTTP headers.
        #! This is used only to show headers to the client.
        self.http_exposed_headers = meta.get('http_exposed_headers')
        if self.http_exposed_headers is None:
            self.http_exposed_headers = (
                'Content-Type',
                'Authorization',
                'Accept',
                'Origin'
            )

        #! List of allowed HTTP origins.
        #! This is used to request or prevent CORS requests.
        #! No CORS requests will be allowed, at-all, unless this
        #! property is set.
        #! NOTE: This can be set to '*' to indicate any origin.
        self.http_allowed_origins = meta.get('http_allowed_origins')
        if self.http_allowed_origins is None:
            self.http_allowed_origins = ()

        #! Whether to use legacy redirects or not to inform the
        #! client the resource is available elsewhere. Legacy redirects
        #! require a combination of 301 and 307 in which 307 is not cacheable.
        #! Modern redirecting uses 308 and is in effect 307 with cacheing.
        #! Unfortunately unknown 3xx codes are treated as
        #! a 300 (Multiple choices) in which the user is supposed to chose
        #! the alternate link so the client is not supposed to auto follow
        #! redirects. Ensure all supported clients understand 308 before
        #! turning off legacy redirecting.
        #! As of 19 March 2013 only Firefox supports it since a year ago.
        self.legacy_redirect = meta.get('legacy_redirect', True)

        #! Mapping of serializers known by this resource.
        #! Values may either be a string reference to the serializer type
        #! or an serializer class object.
        self.serializers = serializers = meta.get('serializers')
        if not serializers:
            self.serializers = {
                'json': 'armet.serializers.JSONSerializer',
                'url': 'armet.serializers.URLSerializer'
            }

        # Expand the serializer name references.
        for name, serializer in six.iteritems(self.serializers):
            if isinstance(serializer, six.string_types):
                segments = serializer.split('.')
                module = '.'.join(segments[:-1])
                module = import_module(module)
                self.serializers[name] = getattr(module, segments[-1])

        #! List of allowed serializers of the understood serializers.
        self.allowed_serializers = meta.get('allowed_serializers')
        if not self.allowed_serializers:
            self.allowed_serializers = tuple(self.serializers.keys())

        # Check to ensure that all allowed serializers are
        # understood serializers.
        for name in self.allowed_serializers:
            if name not in self.serializers:
                raise ImproperlyConfigured(
                    'The allowed serializer, {}, is not one of the '
                    'understood serializers'.format(name))

        #! Name of the default serializer of the list of
        #! understood serializers.
        self.default_serializer = meta.get('default_serializer')
        if not self.default_serializer:
            if 'json' in self.allowed_serializers:
                self.default_serializer = 'json'

            else:
                self.default_serializer = self.allowed_serializers[0]

        if self.default_serializer not in self.allowed_serializers:
            raise ImproperlyConfigured(
                'The chosen default serializer, {}, is not one of the '
                'allowed serializers'.format(self.default_serializer))

        #! Mapping of deserializers known by this resource.
        #! Values may either be a string reference to the deserializer type
        #! or an deserializer class object.
        self.deserializers = deserializers = meta.get('deserializers')
        if not deserializers:
            self.deserializers = {
                'json': 'armet.deserializers.JSONDeserializer',
                'url': 'armet.deserializers.URLDeserializer'
            }

        # Expand the deserializer name references.
        for name, deserializer in six.iteritems(self.deserializers):
            if isinstance(deserializer, six.string_types):
                segments = deserializer.split('.')
                module = '.'.join(segments[:-1])
                module = import_module(module)
                self.deserializers[name] = getattr(module, segments[-1])

        #! List of allowed deserializers of the understood deserializers.
        self.allowed_deserializers = meta.get('allowed_deserializers')
        if not self.allowed_deserializers:
            self.allowed_deserializers = tuple(self.deserializers.keys())

        # Check to ensure that all allowed deserializers are
        # understood deserializers.
        for name in self.allowed_deserializers:
            if name not in self.deserializers:
                raise ImproperlyConfigured(
                    'The allowed deserializer, {}, is not one of the '
                    'understood deserializers'.format(name))

        #! List of authentication protocols to attempt in sequence
        #! to determine the authenticated user.
        self.authentication = meta.get('authentication')
        if self.authentication is None:
            # Default to a single instance of pass-through authentication.
            self.authentication = (authentication.Authentication(),)

        #! The authorization protocol to attempt to use
        #! to determine the if the user can access or is otherwise
        #! authorized.
        self.authorization = meta.get('authorization')
        if self.authorization is None:
            # Default is the pass-through authorization.
            self.authorization = authorization.Authorization()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .parser import (Query, QuerySegment, NoopQuerySegment,
                     BinarySegmentCombinator, UnarySegmentCombinator)

__all__ = [
    'Query',
    'QuerySegment',
    'NoopQuerySegment',
    'BinarySegmentCombinator',
    'UnarySegmentCombinator',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from django.http import HttpResponse
from armet import http
from io import BytesIO
import re


# For almost all headers, django prefixes the header with `HTTP_`.  This is a
# list of headers that are an exception to that rule.
SPECIAL_HEADERS = ('CONTENT_TYPE', 'CONTENT_LENGTH')


def _normalize(name):
    name = re.sub(r'^HTTP-', '', name.replace('_', '-'))
    name = http.request.Headers.normalize(name)
    return name


def _denormalize(name):
    name = name.replace('-', '_').upper()
    return name if name in SPECIAL_HEADERS else 'HTTP_' + name


class RequestHeaders(http.request.Headers):

    def __init__(self, handle):
        #! Reference to the underlying response handle.
        self._handle = handle

        # Continue the initialization.
        super(RequestHeaders, self).__init__()

    @staticmethod
    def normalize(name):
        # Proxy for internal usage of this.
        return _normalize(name)

    def __getitem__(self, name):
        return self._handle.META[_denormalize(name)]

    def __iter__(self):
        for name in self._handle.META:
            if name.startswith('HTTP_') or name in SPECIAL_HEADERS:
                yield self.normalize(name)

    def __len__(self):
        return sum(1 for x in self)

    def __contains__(self, name):
        return _denormalize(name) in self._handle.META


class Request(http.Request):

    def __init__(self, request, *args, **kwargs):
        # Store the request handle.
        self._handle = request

        # Initialize the request headers.
        self.headers = RequestHeaders(request)

        # Store a stream of the request.
        self._stream = BytesIO(request.body)

        # Set the method of the request.
        kwargs.update(method=self._handle.method)

        # Continue the initialization.
        super(Request, self).__init__(*args, **kwargs)

    def _read(self):
        return self._stream.read()

    @property
    def protocol(self):
        return self._handle.META['SERVER_PROTOCOL'].split('/')[0].upper()

    @property
    def mount_point(self):
        path = self._handle.path
        return path[:path.rfind(self.path)] if self.path else path

    @property
    def query(self):
        return self._handle.META.get('QUERY_STRING', '')

    @property
    def uri(self):
        return self._handle.build_absolute_uri()


class ResponseHeaders(http.response.Headers):

    def __init__(self, response, handle):
        #! Reference to the response object.
        self._response = response

        #! Reference to the underlying response handle.
        self._handle = handle

        # Continue the initialization.
        super(ResponseHeaders, self).__init__()

    def __getitem__(self, name):
        return self._handle[name]

    def __setitem__(self, name, value):
        self._response.require_open()
        self._handle[self.normalize(name)] = value

    def __delitem__(self, name):
        self._response.require_open()
        del self._handle[name]

    def __contains__(self, name):
        return self._handle.has_header(name)

    def __iter__(self):
        for name, _ in self._handle.items():
            yield name

    def __len__(self):
        return len(self._handle._headers)


class Response(http.Response):

    def __init__(self, *args, **kwargs):
        # Construct and store a new response object.
        self._handle = HttpResponse()

        # Complete the initialization.
        super(Response, self).__init__(*args, **kwargs)

        # If we're dealing with an asynchronous response, we need
        # to have an asynchronous queue to give to WSGI.
        if self.asynchronous:
            from gevent import queue
            self._queue = queue.Queue()

        # Initialize the response headers.
        self.headers = ResponseHeaders(self, self._handle)

    def __iter__(self):
        # Return the asynchronous queue.
        return self._queue

    @property
    def status(self):
        return self._handle.status_code

    @status.setter
    def status(self, value):
        self.require_open()
        self._handle.status_code = value

    @http.Response.body.setter
    def body(self, value):
        if value:
            if self.asynchronous:
                # Unset the underlying store.
                super(Response, Response).body.__set__(self, None)

                # Write the chunk to the asynchronous queue.
                self._queue.put(value)
                return

        # Set the underlying store.
        super(Response, Response).body.__set__(self, value)

    def close(self):
        # Perform general clean-up and a final flush.
        super(Response, self).close()

        if self.asynchronous:
            # Close the asynchronous queue and terminate the connection
            # to the client.
            self._queue.put(StopIteration)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .resource import Resource
from .managed import ManagedResource
from .model import ModelResource

__all__ = [
    'Resource',
    'ManagedResource',
    'ModelResource'
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet import connectors


def pytest_generate_tests(metafunc):
    # Build a collection of connector hashes to represent every possible
    # arrangement.
    http = connectors.http
    model = connectors.model
    scenarios = [[{'http': x, 'model': y}] for x in http for y in model]
    ids = ['{}:{}'.format(x, y) for x in http for y in model]

    # Parameterize all test classes.
    metafunc.parametrize(['connectors'], scenarios, ids=ids, scope="class")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .attribute import Attribute
from armet import exceptions
import uuid


try:
    import shortuuid

except ImportError:
    shortuuid = None


class UUIDAttribute(Attribute):

    type = uuid.UUID

    def __init__(self, *args, **kwargs):
        super(UUIDAttribute, self).__init__(*args, **kwargs)

        #! Encode and decode the UUID using 'shortuuid' (only if available).
        self.short = kwargs.get('short')
        if self.short is None or self.short:
            if shortuuid is None:
                if self.short is None:
                    self.short = False
                    return

                else:
                    raise exceptions.ImproperlyConfigured(
                        "Use of 'short' UUID attributes requires the "
                        "'shortuuid' package.")

            # Default to using short UUIDs if we have the package.
            self.short = True

    def prepare(self, value):
        if value is None:
            return None

        if self.short:
            # Serialize as a 22-digit hex representation.
            return shortuuid.encode(value)

        # Serialize as the 32-digit hex representation.
        return value.hex

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        try:
            try:
                if self.short:
                    # Attempt to coerce the short UUID.
                    return shortuuid.decode(value)

            except ValueError:
                pass

            # Attempt to coerce the UUID.
            return uuid.UUID(value)

        except ValueError:
            raise ValueError(
                'UUID must be of the form: '
                'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet.http import client


class BaseHTTPException(BaseException):
    status = None

    def __init__(self, content=None, headers=None):
        #! Body of the exception message.
        self.content = content

        #! Additional headers to place with the response.
        self.headers = headers or {}


class BadRequest(BaseHTTPException):  # 400
    status = client.BAD_REQUEST


class Unauthorized(BaseHTTPException):  # 401
    status = client.UNAUTHORIZED


class PaymentRequired(BaseHTTPException):  # 402
    status = client.PAYMENT_REQUIRED


class Forbidden(BaseHTTPException):  # 403
    status = client.FORBIDDEN


class NotFound(BaseHTTPException):  # 404
    status = client.NOT_FOUND


class MethodNotAllowed(BaseHTTPException):  # 405
    status = client.METHOD_NOT_ALLOWED

    def __init__(self, allowed):
        super(MethodNotAllowed, self).__init__(
            headers={'Allow': ', '.join(allowed)})


class NotAcceptable(BaseHTTPException):  # 406
    status = client.NOT_ACCEPTABLE


class ProxyAuthenticationRequired(BaseHTTPException):  # 407
    status = client.PROXY_AUTHENTICATION_REQUIRED


class RequestTimeout(BaseHTTPException):  # 408
    status = client.REQUEST_TIMEOUT


class Conflict(BaseHTTPException):  # 409
    status = client.CONFLICT


class Gone(BaseHTTPException):  # 410
    status = client.GONE


class LengthRequired(BaseHTTPException):  # 411
    status = client.LENGTH_REQUIRED


class PreconditionFailed(BaseHTTPException):  # 412
    status = client.PRECONDITION_FAILED


class RequestEntityTooLarge(BaseHTTPException):  # 413
    status = client.REQUEST_ENTITY_TOO_LARGE


class RequestUriTooLong(BaseHTTPException):  # 414
    status = client.REQUEST_URI_TOO_LONG


class UnsupportedMediaType(BaseHTTPException):  # 415
    status = client.UNSUPPORTED_MEDIA_TYPE


class RequestedRangeNotSatisfiable(BaseHTTPException):  # 416
    status = client.REQUESTED_RANGE_NOT_SATISFIABLE


class ExpectationFailed(BaseHTTPException):  # 417
    status = client.EXPECTATION_FAILED


class UnprocessableEntity(BaseHTTPException):  # 422
    status = client.UNPROCESSABLE_ENTITY


class Locked(BaseHTTPException):  # 423
    status = client.LOCKED


class FailedDependency(BaseHTTPException):  # 424
    status = client.FAILED_DEPENDENCY


class UgradeRequired(BaseHTTPException):  # 426
    status = client.UPGRADE_REQUIRED


class InternalServerError(BaseHTTPException):  # 500
    status = client.INTERNAL_SERVER_ERROR


class NotImplemented(BaseHTTPException):  # 501
    status = client.NOT_IMPLEMENTED


class BadGateway(BaseHTTPException):  # 502
    status = client.BAD_GATEWAY


class ServiceUnavailable(BaseHTTPException):  # 503
    status = client.SERVICE_UNAVAILABLE


class GatewayTimeout(BaseHTTPException):  # 504
    status = client.GATEWAY_TIMEOUT


class HttpVersionNotSupported(BaseHTTPException):  # 505
    status = client.HTTP_VERSION_NOT_SUPPORTED


class InsufficientStorage(BaseHTTPException):  # 507
    status = client.INSUFFICIENT_STORAGE


class NotExtended(BaseHTTPException):  # 510
    status = client.NOT_EXTENDED



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import json
from armet import http
from .base import BaseResourceTest


class TestResourceValidation(BaseResourceTest):

    def test_post_too_low(self, connectors):
        data = {'votes': -14}
        body = json.dumps(data)
        response, content = self.client.post(
            path='/api/poll-valid/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert 'votes' in data
        assert data['votes'] == ['Must be greater than 0.']

    def test_post_too_high(self, connectors):
        data = {'votes': 54}
        body = json.dumps(data)
        response, content = self.client.post(
            path='/api/poll-valid/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert 'votes' in data
        assert data['votes'] == ['Must be less than 51.']

    def test_post_lots_wrong(self, connectors):
        data = {'votes': 0, 'question': 'This'}
        body = json.dumps(data)
        response, content = self.client.post(
            path='/api/poll-valid/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert 'votes' in data
        assert 'question' in data
        assert data['votes'][0] == 'Must be greater than 0.'
        assert data['question'][0] == 'Must be more than 15 characters.'
        assert data['question'][1] == 'Must have at least one question mark.'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import unittest
import six
from armet.query import parser, constants
from pytest import mark


@mark.bench('parser.parse', iterations=10000)
class QueryTestCase(unittest.TestCase):

    def parse(self, text):
        """Simple convenience function to unwrap the array of parameters."""
        return parser.parse(text).parsed

    def test_empty(self):
        """Test an empty query string."""
        item = self.parse('')

        assert isinstance(item, parser.NoopQuerySegment)

    def test_simple_filter(self):
        item = self.parse('foo=bar')

        assert item.path == ['foo']
        assert item.operator, constants.OPERATOR_IEQUAL[0]
        assert not item.negated
        assert item.values == ['bar']

    def test_binary(self):
        item = self.parse(b'foo=bar')

        assert item.path == ['foo']
        assert item.values == ['bar']

    def test_negation(self):
        queries = ['foo!=bar', 'foo.not=bar']
        for query in queries:
            item = self.parse(query)

            assert item.negated

    def test_relational_filter(self):
        item = self.parse('bread.sticks=delicious')

        assert item.path == ['bread', 'sticks']
        assert item.operator, constants.OPERATOR_IEQUAL[0]
        assert not item.negated
        assert item.values == ['delicious']

    def test_values(self):
        item = self.parse('fruit=apples,oranges')

        assert item.path == ['fruit']
        assert item.operator, constants.OPERATOR_IEQUAL[0]
        assert not item.negated
        assert item.values == ['apples', 'oranges']

    def test_bogus(self):
        """Test some bogusy query strings."""
        queries = [
            'foo:asc&;bar:desc',
            'foo.lte<=3',
            'foo.!negate',
            'lte=3',
        ]
        for query in queries:
            self.assertRaises(ValueError, self.parse, query)

    def test_operations(self):
        for name, symbol in constants.OPERATORS:
            item = self.parse('crazy.{}=true'.format(name))

            assert item.path == ['crazy']
            assert item.operator == constants.OPERATOR_SUFFIX_MAP[name]
            assert not item.negated
            assert item.values == ['true']

            if symbol is not None:
                item = self.parse('crazy{}true'.format(symbol))

                assert item.path == ['crazy']
                assert item.operator == constants.OPERATOR_EQUALITY_MAP[symbol]
                assert not item.negated
                assert item.values == ['true']

    def test_fusion(self):
        """Test something from everything combined"""
        q = ('the.rolling.stones.iregex.not:asc=sympathy,for,the,devil;'
             '!(guns.n.roses=paradise,city&queen:asc)')

        item = self.parse(q)

        # I'm lazy and don't want to walk the tree, so lets just test the repr.
        assert (six.text_type(repr(item)) ==
                "(the.rolling.stones.iregex.not:asc :iexact 'sympathy' "
                "| 'for' | 'the' | 'devil') OR NOT (guns.n.roses :iexact "
                "'paradise' | 'city') AND (queen :iexact '')")

    def test_grouping(self):
        item = self.parse('foo=bar&(a=b;b=c)')

        assert item.left.path == ['foo']
        assert item.right.left.path == ['a']
        assert item.right.right.path == ['b']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import sys
from six.moves import reload_module
from importlib import import_module


def force_import_module(name):
    if name in sys.modules:
        return reload_module(sys.modules[name])

    else:
        return import_module(name)


def unload_module(name):
    if name in sys.modules:
        del sys.modules[name]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import collections
import abc
import six
import mimeparse
import weakref
import io
from armet import exceptions
from . import request, client


class Headers(collections.MutableMapping, request.Headers):
    """Describes a mutable mapping abstraction over response headers.
    """

    class _Header(collections.MutableSequence):
        """
        Provides an implementation of a mutable sequence of multi-valued
        headers that are synced with the normal list of headers.
        """

        def __init__(self, sequence, name):
            #! Internal reference to the first-degree headers dictionary.
            self._headers = sequence._headers
            self._obj = self._headers._obj
            self._name = name

            #! Internal store of the value at instantiation of the header.
            self._value = self._headers.get(self._name, '')

            #! Internal sequence of the list of the values of the header.
            self._sequence = self._value.split(',')

        def _pull(self):
            # Pull the headers from the source.
            if self._value != self._headers.get(self._name, ''):
                # If the header value has changed; update our sequence.
                self._value = self._headers.get(self._name, '')
                self._sequence = self._value.split(',')

        def __getitem__(self, index):
            # Retrieve the header value at the passed index.
            self._pull()
            return self._sequence[index]

        def __len__(self):
            self._pull()
            return len(self._sequence)

        def _push(self):
            # Push the headers to the source.
            self._value = ','.join(self._sequence)
            self._headers[self._name] = self._value

        def __setitem__(self, index, value):
            # Store the value at the passed index
            self._obj._assert_open()
            self._sequence[index] = value
            self._push()

        def __delitem__(self, index):
            # Remove the value at the passed index.
            self._obj._assert_open()
            del self._sequence[index]
            self._push()

        def insert(self, index, value):
            # Insert the header value at the passed index.
            self._obj._assert_open()
            self._sequence.insert(index, value)
            self._push()

    @abc.abstractmethod
    def __setitem__(self, name, value):
        """Stores a header with the passed name.

        @param[in] name
            The name to store the header as. This is passed through
            `Headers.normalize` before storing on the response.

        @param[in] value
            The value to store for the header; for multi-valued headers,
            this can be a comma-separated list of values.
        """

    @abc.abstractmethod
    def __delitem__(self, name):
        """Removes a header with the passed name.

        @param[in] name
            The case-insensitive name of the header to remove
            from the response.
        """

    def append(self, name, value):
        """Add a value to the end of the list for the named header."""
        return self._sequence[name].append(value)

    def extend(self, name, values):
        """Extend the list for the named header by appending all values."""
        return self._sequence[name].extend(values)

    def insert(self, name, index, value):
        """Insert a value at the passed index in the named header."""
        return self._sequence[name].insert(index, value)

    def remove(self, name, value):
        """
        Remove the first item with the passed value from the
        list for the named header.
        """
        return self._sequence[name].remove(value)

    def popvalue(self, name, index=None):
        """Remove the item at the given position in the named header list."""
        return self._sequence[name].pop(index)

    def sort(self, name):
        """Sort the items of the list, in place."""
        return self._sequence[name].sort()

    def reverse(self, name):
        """Reverse the elements of the list, in place."""
        return self._sequence[name].reverse()


class Response(object):
    """Describes the RESTful response abstraction.
    """

    #! Dictionary-like interface to access headers; this should be
    #! set by the dervied class to an instance of a derived Headers class.
    headers = None

    def __init__(self, asynchronous, *args, **kwargs):
        #! True if the response object is closed.
        self._closed = False

        #! True if the response object is streaming.
        self.streaming = False

        #! True if we're asynchronous.
        self.asynchronous = asynchronous

        #! Default the status code to OK.
        self.status = client.OK

        #! A reference to the bound resource; this is set in the resource
        #! view method after traversal.
        self._resource = None

        #! Explicit declaration of character encoding to use when
        #! writing data to the response body.
        #! Defaults to parsing content-type and determining encoding
        #! via the standard rules.
        self._encoding = None

        #! The underlying file stream to write incoming data to.
        self._stream = io.BytesIO()

        #! The content chunk to return to the client.
        self._body = None

        #! The length of the response.
        self._length = 0

    def require_not_closed(self):
        """Raises an exception if the response is closed."""
        if self.closed:
            raise exceptions.InvalidOperation('Response is closed.')

    def require_open(self):
        """Raises an exception if the response is not open."""
        self.require_not_closed()
        if self.streaming:
            raise exceptions.InvalidOperation('Response is streaming.')

    @property
    def status(self):
        """Gets the status code of the response."""
        raise NotImplementedError()

    @status.setter
    def status(self, value):
        """Sets the status code of the response."""
        raise NotImplementedError()

    @property
    def body(self):
        """Returns the current value of the response body."""
        return self._body

    @body.setter
    def body(self, value):
        """Sets the response body to the passed value.

        @note
            During asynchronous or streaming responses, remember that
            the `body` property refers to the portion of the response *not*
            sent to the client.
        """
        self._body = value

    def bind(self, resource):
        """Binds this to the passed resource object.

        @sa armet.http.request.Request.bind
        """
        self._resource = weakref.proxy(resource)

    @property
    def resource(self):
        return self._resource

    @property
    def encoding(self):
        if self._encoding is not None:
            # Encoding has been set manually.
            return self._encoding

        # Get the `Content-Type` header, if available.
        content_type = self.headers.get('Content-Type')
        if content_type:
            # Parse out the primary type and parameters from the media type.
            ptype, _, params = mimeparse.parse_mime_type(content_type)

            # Return the specified charset or the default depending on the
            # primary type.
            default = 'utf-8' if ptype == 'application' else 'iso-8859-1'
            return params.get('charset', default)

        # No encoding found.

    @encoding.setter
    def encoding(self, value):
        # Explicitly set the character encoding to use.
        self._encoding = value

    @abc.abstractmethod
    def close(self):
        """Flush and close the stream.

        This is called automatically by the base resource on resources
        unless the resource is operating asynchronously; in that case,
        this method MUST be called in order to signal the end of the request.
        If not the request will simply hang as it is waiting for some
        thread to tell it to return to the client.
        """

        # Ensure we're not closed.
        self.require_not_closed()

        if not self.streaming or self.asynchronous:
            # We're not streaming, auto-write content-length if not
            # already set.
            if 'Content-Length' not in self.headers:
                self.headers['Content-Length'] = self.tell()

        # Flush out the current buffer.
        self.flush()

        # We're done with the response; inform the HTTP connector
        # to close the response stream.
        self._closed = True

    @property
    def closed(self):
        """True if the stream is closed."""
        return self._closed

    def tell(self):
        """Return the current stream position."""
        return self._length

    def write(self, chunk, serialize=False, format=None):
        """Writes the given chunk to the output buffer.

        @param[in] chunk
            Either a byte array, a unicode string, or a generator. If `chunk`
            is a generator then calling `self.write(<generator>)` is
            equivalent to:

            @code
                for x in <generator>:
                    self.write(x)
                    self.flush()
            @endcode

        @param[in] serialize
            True to serialize the lines in a determined serializer.

        @param[in] format
            A specific format to serialize in; if provided, no detection is
            done. If not provided, the accept header (as well as the URL
            extension) is looked at to determine an appropriate serializer.
        """

        # Ensure we're not closed.
        self.require_not_closed()

        if chunk is None:
            # There is nothing here.
            return

        if serialize or format is not None:
            # Forward to the serializer to serialize the chunk
            # before it gets written to the response.
            self.serialize(chunk, format=format)
            return  # `serialize` invokes write(...)

        if type(chunk) is six.binary_type:
            # Update the stream length.
            self._length += len(chunk)

            # If passed a byte string, we hope the user encoded it properly.
            self._stream.write(chunk)

        elif isinstance(chunk, six.string_types):
            encoding = self.encoding
            if encoding is not None:
                # If passed a string, we can encode it for the user.
                chunk = chunk.encode(encoding)

            else:
                # Bail; we don't have an encoding.
                raise exceptions.InvalidOperation(
                    'Attempting to write textual data without an encoding.')

            # Update the stream length.
            self._length += len(chunk)

            # Write the encoded data into the byte stream.
            self._stream.write(chunk)

        elif isinstance(chunk, collections.Iterable):
            # If passed some kind of iterator, attempt to recurse into
            # oblivion.
            for section in chunk:
                self.write(section)

        else:
            # Bail; we have no idea what to do with this.
            raise exceptions.InvalidOperation(
                'Attempting to write something not recognized.')

    def serialize(self, data, format=None):
        """Serializes the data into this response using a serializer.

        @param[in] data
            The data to be serialized.

        @param[in] format
            A specific format to serialize in; if provided, no detection is
            done. If not provided, the accept header (as well as the URL
            extension) is looked at to determine an appropriate serializer.

        @returns
            A tuple of the serialized text and an instance of the
            serializer used.
        """
        return self._resource.serialize(data, response=self, format=format)

    def flush(self):
        """Flush the write buffers of the stream.

        This results in writing the current contents of the write buffer to
        the transport layer, initiating the HTTP/1.1 response. This initiates
        a streaming response. If the `Content-Length` header is not given
        then the chunked `Transfer-Encoding` is applied.
        """

        # Ensure we're not closed.
        self.require_not_closed()

        # Pull out the accumulated chunk.
        chunk = self._stream.getvalue()
        self._stream.truncate(0)
        self._stream.seek(0)

        # Append the chunk to the body.
        self.body = chunk if (self._body is None) else (self._body + chunk)

        if self.asynchronous:
            # We are now streaming because we're asynchronous.
            self.streaming = True

    def send(self, *args, **kwargs):
        """Writes the passed chunk and flushes it to the client."""
        self.write(*args, **kwargs)
        self.flush()

    def end(self, *args, **kwargs):
        """
        Writes the passed chunk, flushes it to the client,
        and terminates the connection.
        """
        self.send(*args, **kwargs)
        self.close()

    def __getitem__(self, name):
        """Retrieves a header with the passed name."""
        return self.headers[name]

    def __setitem__(self, name, value):
        """Stores a header with the passed name."""
        self.headers[name] = value

    def __delitem__(self, name):
        """Removes a header with the passed name."""
        del self.headers

    def __len__(self):
        """Retrieves the actual length of the response."""
        return self.tell()

    def __nonzero__(self):
        """Test if the response is closed."""
        return not self._closed

    def __bool__(self):
        """Test if the response is closed."""
        return not self._closed

    def __contains__(self, name):
        """Tests if the passed header exists in the response."""
        return name in self.headers

    def append(self, name, value):
        """Add a value to the end of the list for the named header."""
        return self.headers.append(name, value)

    def extend(self, name, values):
        """Extend the list for the named header by appending all values."""
        return self.headers.extend(name, values)

    def insert(self, name, index, value):
        """Insert a value at the passed index in the named header."""
        return self.headers.insert(index, value)

    def remove(self, name, value):
        """
        Remove the first item with the passed value from the
        list for the named header.
        """
        return self.headers.remove(name, value)

    def popvalue(self, name, index=None):
        """Remove the item at the given position in the named header list."""
        return self.headers.popvalue(name, index)

    def index(self, name, value):
        """
        Return the index in the list of the first item whose value is x in
        the values of the named header.
        """
        return self.headers.index(name, value)

    def count(self, name, value):
        """
        Return the number of times a value appears in the list of the values
        of the named header.
        """
        return self.headers.count(name, value)

    def sort(self, name):
        """Sort the items of the list, in place."""
        return self.headers.sort(name)

    def reverse(self, name):
        """Reverse the elements of the list, in place."""
        return self.headers.reverse(name)

    def getlist(self, name):
        """Retrieves the passed header as a sequence of its values."""
        return self.headers.getlist(name)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet.utils import compose
from collections import defaultdict
from functools import partial


class Attribute(object):
    """Generic attribute.

    Use this to indicate that the type of the data is unmanaged by armet.
    This is also an extension point for further attributes.
    """

    #! Python type expected to be marshalled through this attribute.
    type = None

    def __init__(self, path=None, **kwargs):

        #! Attribute may be accessed directly.
        #! Note that setting this to false will still show the attribute in
        #! the body. Set include to false to hide it from the body.
        self.read = kwargs.get('read', True)

        #! Attribute may be modified through any operation.
        self.write = kwargs.get('write', True)

        #! Attribute is included in the response body.
        self.include = kwargs.get('include', True)

        #! Attribute can accept a null value.
        self.null = kwargs.get('null', True)

        #! Attribute must be present in the body.
        self.required = kwargs.get('required', False)

        #! Attribute is to be treated as a collection.
        self.collection = kwargs.get('collection', False)

        #! Override python name of the attribute.
        self.name = kwargs.get('name')

        #! Flag to turn the set operation into a no-op.
        # HACK: This is in-place in order to get around a bug I've encountered
        #   until relationships are implemented.
        self._set = kwargs.get('_set', True)

        #! The path reference of where to find this attribute on an
        #! item (eg. 'name' references the name key if the read method returns
        #! a dictionary.)
        #!
        #! The path may be dot-separated to indicate simple traversal.
        #! Eg. user.name could be obj['user'].name
        self.path = path
        if self.path:
            # Explode the path into segments to iterate over in operations.
            self._segments = defaultdict(lambda: self.path.split('.'))

        # Attributes use a lazy optimization process for attribute lookup
        # on the underlying data model. When an attribute is accessed
        # it builds and caches the operation action.
        self._getters = defaultdict(list)

        # Attribute cache that is keyed on the type of the target object.
        this = self

        class GetResolver(dict):

            def __missing__(self, key):
                return partial(this._resolve_get, self, key.__class__)

        self._get = GetResolver()

    def _resolve_get(self, resolver, key, target):
        if self.path is None:
            # There is no path defined on this resource.
            # We can do no magic to get the value.
            resolver[key] = lambda *a: None
            return None

        # Iterate and resolve each constructed getter.
        for func in self._getters[key]:
            if target is None:
                # No more getters can be resolved.
                return None

            # Resolve this getter and continue iteration.
            target = func(target)

        # Are their segments remaining to make a getter for?
        while self._segments[key]:
            if target is None:
                # No more getters can be made.
                return None

            # Fetch the next path segment.
            segment = self._segments[key].pop(0)

            # Make and append the corresponding getter.
            func = self._make_getter(segment, key)
            self._getters[key].append(func)

            # Resolve the getter.
            target = func(target)

        # Have we finished resolving the getters?
        if not self._segments[key]:
            # Compose a replacement get function that just iterates over
            # the getters.
            resolver[key] = compose(*self._getters[key])

        # Return our resolved value.
        return target

    def get(self, target):
        """Retrieve the value of this attribute from the passed object.
        """

        # Attempt to resolve an accessor for the specific target. Creates
        # the accessor if not available.
        return self._get[target](target)

    def set(self, target, value):
        """Set the value of this attribute for the passed object.
        """

        if not self._set:
            return

        if self.path is None:
            # There is no path defined on this resource.
            # We can do no magic to set the value.
            self.set = lambda *a: None
            return None

        if self._segments[target.__class__]:
            # Attempt to resolve access to this attribute.
            self.get(target)

        if self._segments[target.__class__]:
            # Attribute is not fully resolved; an interim segment is null.
            return

        # Resolve access to the parent object.
        # For a single-segment path this will effectively be a no-op.
        parent_getter = compose(*self._getters[target.__class__][:-1])
        target = parent_getter(target)

        # Make the setter.
        func = self._make_setter(self.path.split('.')[-1], target.__class__)

        # Apply the setter now.
        func(target, value)

        # Replace this function with the constructed setter.
        def setter(target, value):
            func(parent_getter(target), value)

        self.set = setter

    def prepare(self, value):
        """Prepare the value for serialization and presentation to the client.
        """

        # By default, do nothing to the value.
        return value

    def clean(self, value):
        """Cleans the value from deserialization into consumption by python.
        """

        # By default, do nothing to the value.
        return value

    def try_clean(self, value):
        try:
            # TODO: Remove usage of this in query builder.
            return self.clean(value)

        except (ValueError, AssertionError):
            return None

    def clone(self):
        """Construct an identical attribute.

        Used by the resource metaclass to ensure all attributes are unique
        instances. This is done so that when getters and setters are resolved
        the caches don't clobber base classes (for inherited attributes).
        """

        return self.__class__(**self.__dict__)

    def _make_getter(self, segment, class_):

        # Attempt to resolve properties and simple functions by
        # accessing the class attribute directly.
        obj = getattr(class_, segment, None)
        if obj is not None:
            if hasattr(obj, '__call__'):
                return obj.__call__

            if hasattr(obj, '__get__'):
                return lambda target, x=obj.__get__: x(target, class_)

        # Check for much better hidden descriptor.
        obj = class_.__dict__.get(segment)
        if obj is not None and hasattr(obj, '__get__'):
            return lambda target, x=obj.__get__: x(target, class_)

        # Check for item access (for a dictionary).
        if hasattr(class_, '__getitem__'):
            def getter(target):
                try:
                    return target[segment]

                except KeyError:
                    return None

            return getter

        # Check for attribute access.
        if hasattr(class_, '__dict__'):
            return lambda target: target.__dict__.get(segment)

        raise RuntimeError(
            'unable to resolve attribute access for %r on %r' % (
                segment, class_))

    def _make_setter(self, segment, class_):

        # Attempt to resolve a data descriptor.
        obj = getattr(class_, segment, None)
        if obj is not None:
            if hasattr(obj, '__set__'):
                def setter(target, value, x=obj.__set__):
                    x(target, value)

                return setter

        # Check for much better hidden descriptor.
        obj = class_.__dict__.get(segment)
        if obj is not None and hasattr(obj, '__set__'):
            def setter(target, value, x=obj.__set__):
                x(target, value)

            return setter

        # Check for item access (for a dictionary).
        if hasattr(class_, '__getitem__'):
            def setter(target, value):
                target[segment] = value

            return setter

        # Check for attribute access.
        if hasattr(class_, '__dict__'):
            def setter(target, value):
                target.__dict__[segment] = value

            return setter

        raise RuntimeError(
            'unable to resolve attribute setter for %r on %r' % (
                segment, class_))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import operator
import re


#! Equality
OPERATOR_EQUAL = 'exact', '=='

#! Case-insensitve equality
OPERATOR_IEQUAL = 'iexact', '='

#! Less than
OPERATOR_LT = 'lt', '<'

#! Less than or equal to
OPERATOR_LTE = 'lte', '<='

#! Greater than
OPERATOR_GT = 'gt', '>'

#! Greater than or equal to
OPERATOR_GTE = 'gte', '>='

#! Regular expression match
OPERATOR_REGEX = 'regex', '*='

#! Null test
OPERATOR_ISNULL = 'isnull', None

#! Contains (x in y)
OPERATOR_ICONTAINS = 'icontains', None

#! The fallback to use in the case that a more specific one isn't defined.
OPERATOR_FALLBACK = OPERATOR_IEQUAL
OPERATOR_SUFFIX_FALLBACK = OPERATOR_FALLBACK[0]
OPERATOR_EQUALITY_FALLBACK = OPERATOR_FALLBACK[1]

#! Operator map relating operations to python operations.
# TODO: the values of the operator map are somewhat unwieldy to work with
# perhaps the value stored in a Segment() type should instead be one of
# the keys instead of the values.  The values just show a convenient thing
# to work with in pythonland.
OPERATOR_MAP = {
    OPERATOR_EQUAL: operator.eq,
    OPERATOR_IEQUAL: lambda x, y: x.lower() == y.lower(),
    OPERATOR_LT: operator.lt,
    OPERATOR_LTE: operator.le,
    OPERATOR_GT: operator.gt,
    OPERATOR_GTE: operator.ge,
    OPERATOR_REGEX: lambda x, y: re.search(y, x),
    OPERATOR_ISNULL: lambda x: x is None,
    OPERATOR_ICONTAINS: operator.contains,
}

#! Operator set containing all operators.
OPERATORS = set(OPERATOR_MAP.keys())

#! Operators restricted to suffixed paths (foo.gte=bar)
OPERATOR_SUFFIXES = set(filter(None, map(operator.itemgetter(0), OPERATORS)))

#! Operators restricted to equality comparators (foo>=bar)
OPERATOR_EQUALITIES = set(filter(None, map(operator.itemgetter(1), OPERATORS)))

#! Operator map limited to suffixed paths {"gte": operator.ge}
OPERATOR_SUFFIX_MAP = dict(
    filter(operator.itemgetter(0),
           map(lambda x: (x[0][0], x[1]),
               OPERATOR_MAP.items())))

#! Operator map limited to equality checks {">=": operator.ge}
OPERATOR_EQUALITY_MAP = dict(
    filter(operator.itemgetter(0),
           map(lambda x: (x[0][1], x[1]),
               OPERATOR_MAP.items())))

#! Negation
PATH_NEGATION = 'not'
OPERATOR_NEGATION = '!'

NEGATION = (PATH_NEGATION, OPERATOR_NEGATION)

#! Logical
LOGICAL_AND = '&'
LOGICAL_OR = ';'

#! Path separator
SEP_PATH = '.'

#! Value separator
SEP_VALUE = ','

#! Directive initiator
DIRECTIVE = ':'

#! Grouping characters
GROUP_BEGIN = '('
GROUP_END = ')'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class ImproperlyConfigured(BaseException):
    """Something has been set up or configured incorrectly.
    """


class InvalidOperation(BaseException):
    """
    Something is being asked to operate outside of the defined specification
    of it operating.
    """


class ValidationError(BaseException):
    """
    Something has been found invalid during the attribute clean cycle;
    normally resulting from form or some type of validation.
    """

    def __init__(self, *errors):
        self.errors = errors



# -*- coding: utf-8 -*-
from armet import http
from .base import BaseResourceTest


class TestResourceOptions(BaseResourceTest):

    @classmethod
    def setup_class(cls):
        super(TestResourceOptions, cls).setup_class()
        cls.right_path = '/api/left/'
        cls.left_path = '/api/right/'
        cls.right_origin = 'http://127.0.0.1:80'
        cls.left_origin = 'http://10.0.0.1'
        cls.default_method = 'GET'

    def test_origin_header(self, connectors):
        # Check an endpoint with a specific origin.
        # Try with no Origin request.
        # Should just send back 200.
        response, _ = self.client.options(self.right_path)

        assert response.status == http.client.OK

        # Try with Origin not in the list of allowed origins.
        response, _ = self.client.options(
            self.right_path,
            headers={'Origin': self.left_origin})

        assert response.status == http.client.OK

        # Try with the correct Origin.
        response, _ = self.client.options(
            self.right_path,
            headers={'Origin': self.right_origin})

        assert response.status == http.client.OK

        # Check an endpoint which allows all origins.
        # Try with no Origin request.
        # Should just send back 200.
        response, _ = self.client.options(self.left_path)

        assert response.status == http.client.OK

        # Try with any random origin.
        response, _ = self.client.options(
            self.left_path,
            headers={'Origin': self.left_origin})

        assert response.status == http.client.OK

    def access_control_headers(self, header):
        """
        Test each Access-Control-Allow header here, since they
        all do the same thing.
        """

        # Check an endpoint that has a specific origin.
        # Check with no origin.
        response, _ = self.client.options(self.right_path)

        # Should give back a 200 and no Allow-Headers header.
        assert response.status == http.client.OK
        assert header not in response

        # Set the origin to the correct place.
        response, _ = self.client.options(
            self.right_path,
            headers={
                'Origin': self.right_origin,
                'Access-Control-Request-Method': self.default_method})

        # Should give back a 200 and an Allow-Headers header.
        assert response.status == http.client.OK
        assert header not in response

        # Do it again with asterisk.
        response, _ = self.client.options(
            self.right_path, headers={'Origin': '*'})

        # Should give back a 200 and no Allow-Headers header.
        assert response.status == http.client.OK
        assert header not in response

        # TODO: Should probably check that we're getting the
        #       correct methods back.

        # Now try with an enpoint that allows any origin.
        # Check with no origin.
        response, _ = self.client.options(self.left_path)

        # Should give back a 200 and no Allow-Headers header.
        assert response.status == http.client.OK
        assert header not in response

        # Set the origin to the correct place.
        response, _ = self.client.options(
            self.left_path,
            headers={
                'Origin': self.left_origin,
                'Access-Control-Request-Method': self.default_method})

        # Should give back a 200 and no Allow-Headers header.
        assert response.status == http.client.OK
        assert header not in response

        # Should probably check that we're getting the correct methods back.

    def test_access_control_allow_headers(self, connectors):
        """Test the Access-Control-Allow-Headers header."""
        self.access_control_headers('Access-Control-Allow-Headers')

    def test_access_control_allow_methods(self, connectors):
        """Test the Access-Control-Allow-Methods header."""
        self.access_control_headers('Access-Control-Allow-Methods')

    def test_access_control_allow_origin(self, connectors):
        """Test the Access-Control-Allow-Origin header."""
        self.access_control_headers('Access-Control-Allow-Origin')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import unittest
import uuid
from armet import deserializers
import ujson as json
from pytest import mark


class DeserializerTestCase(unittest.TestCase):

    Deserializer = None

    @classmethod
    def setup_class(cls):
        cls.deserializer = cls.Deserializer()

    def deserialize(self, text):
        self.data = self.deserializer.deserialize(text=text)


@mark.bench('self.deserialize', iterations=10000)
class JSONDeserializerTestCase(DeserializerTestCase):

    Deserializer = deserializers.JSONDeserializer

    def test_scalar(self):
        self.assertRaises(ValueError, self.deserialize, 124)

    def test_number(self):
        self.deserialize(b'[42]')

        assert self.data == [42]

    def test_boolean(self):
        self.deserialize(b'[true]')

        assert self.data == [True]

        self.deserialize(b'[false]')

        assert self.data == [False]

    def test_array(self):
        self.deserialize(b'[1, 2, 3]')

        assert self.data == [1, 2, 3]

    def test_dict(self):
        self.deserialize(b'{"x": 2, "y": "bob"}')

        assert self.data == {"x": 2, "y": "bob"}

    def test_dict_array(self):
        self.deserialize(b'{"x": [1, 2], "y": "bob"}')

        assert self.data == {"x": [1, 2], "y": "bob"}

    def test_large(self):
        payload_item = {
            'organization': uuid.uuid4().hex,
            'user': uuid.uuid4().hex,
            'id': uuid.uuid4().hex,
            'is_active': False,
            'is_pending': True,
            # 'members': [
            #     {
            #         'id': uuid.uuid4().hex,
            #         'organization': uuid.uuid4().hex,
            #         'is_dead': False
            #     }
            # ],
        }

        payload = [payload_item] * 100

        text = json.dumps(payload, ensure_ascii=False)

        self.deserialize(text.encode('utf8'))

        assert self.data == payload


@mark.bench('self.deserialize', iterations=10000)
class URLDeserializerTestCase(DeserializerTestCase):

    Deserializer = deserializers.URLDeserializer

    def test_scalar(self):
        self.assertRaises(ValueError, self.deserialize, 124)

    def test_sequence(self):
        self.deserialize(b'x=2&y=51')

        assert self.data == {'x': ['2'], 'y': ['51']}

    def test_multi_sequence(self):
        self.deserialize(b'x=2&y=51&x=781&y=165')

        assert self.data == {"x": ['2', '781'], "y": ['51', '165']}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import collections


def cons(collection, value):
    """Extends a collection with a value."""
    if isinstance(value, collections.Mapping):
        if collection is None:
            collection = {}
        collection.update(**value)

    elif isinstance(value, six.string_types):
        if collection is None:
            collection = []
        collection.append(value)

    elif isinstance(value, collections.Iterable):
        if collection is None:
            collection = []
        collection.extend(value)

    else:
        if collection is None:
            collection = []
        collection.append(value)

    return collection


def compose(*functions):

    def composed(x):
        for func in functions:
            x = func(x)

        return x

    return composed



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import logging
import re
import six
import collections
import mimeparse
from armet import http, utils


logger = logging.getLogger(__name__)


class Resource(object):
    """Implements the RESTful resource protocol for abstract resources.

    @note
        This is not the class to derive from when implementing your own
        resources. Derive from `armet.resources.Resource` (defined in
        the `__init__.py`).
    """

    #! Instantiated options class; contains all merged properties
    #! from the class hierarchy's Meta classes.
    meta = None

    #! Maps media ranges to deserializer names.
    #! Generated by the metaclass.
    _deserializer_map = None

    #! Maps media ranges to serializer names.
    #! Generated by the metaclass.
    _serializer_map = None

    def __new__(cls, request, response, *args, **kwargs):
        # Parse any arguments out of the path and traverse down the
        # path using any defined patterns.
        cls, params = cls.traverse(request)

        # Actually construct the resource and return the instance.
        obj = super(Resource, cls).__new__(cls)

        # Update our instance dictionary with the arugments from `parse`.
        obj.__dict__.update(params)

        # Return our constructed instance.
        return obj

    @classmethod
    def redirect(cls, request, response):
        """Redirect to the canonical URI for this resource."""
        if cls.meta.legacy_redirect:
            if request.method in ('GET', 'HEAD',):
                # A SAFE request is allowed to redirect using a 301
                response.status = http.client.MOVED_PERMANENTLY

            else:
                # All other requests must use a 307
                response.status = http.client.TEMPORARY_REDIRECT

        else:
            # Modern redirects are allowed. Let's have some fun.
            # Hopefully you're client supports this.
            # The RFC explicitly discourages UserAgent sniffing.
            response.status = http.client.PERMANENT_REDIRECT

        # Terminate the connection.
        response.close()

    @classmethod
    def view(cls, request, response):
        """
        Entry-point of the request / response cycle; Handles resource creation
        and delegation.

        @param[in] requset
            The HTTP request object; containing accessors for information
            about the request.

        @param[in] response
            The HTTP response object; contains accessors for modifying
            the information that will be sent to the client.
        """
        # Determine if we need to redirect.
        test = cls.meta.trailing_slash
        if test ^ request.path.endswith('/'):
            # Construct a new URL by removing or adding the trailing slash.
            path = request.path + '/' if test else request.path[:-1]
            response['Location'] = '{}://{}{}{}{}'.format(
                request.protocol.lower(),
                request.host,
                request.mount_point,
                path,
                '?' + request.query if request.query else '')

            # Redirect to the version with the correct trailing slash.
            return cls.redirect(request, response)

        try:
            # Instantiate the resource.
            obj = cls(request, response)

            # Bind the request and response objects to the constructed
            # resource.
            request.bind(obj)
            response.bind(obj)

            # Bind the request object to the resource.
            # This is used to facilitate the serializer and deserializer.
            obj._request = request

            # Initiate the dispatch cycle and handle its result on
            # synchronous requests.
            result = obj.dispatch(request, response)

            if not response.asynchronous:
                # There is several things that dispatch is allowed to return.
                if (isinstance(result, collections.Iterable) and
                        not isinstance(result, six.string_types)):
                    # Return the stream generator.
                    return cls.stream(response, result)

                else:
                    # Leave it up to the response to throw or write whatever
                    # we got back.
                    response.end(result)
                    if response.body:
                        # Return the body if there was any set.
                        return response.body

        except http.exceptions.BaseHTTPException as e:
            # Something that we can handle and return properly happened.
            # Set response properties from the exception.
            response.status = e.status
            response.headers.update(e.headers)

            if e.content:
                # Write the exception body if present and close
                # the response.
                # TODO: Use the plain-text encoder.
                response.send(e.content, serialize=True, format='json')

            # Terminate the connection and return the body.
            response.close()
            if response.body:
                return response.body

        except Exception:
            # Something unexpected happened.
            # Log error message to the logger.
            logger.exception('Internal server error')

            # Write a debug message for the client.
            if not response.streaming and not response.closed:
                response.status = http.client.INTERNAL_SERVER_ERROR
                response.headers.clear()
                response.close()

    @classmethod
    def parse(cls, path):
        """Parses out parameters and separates them out of the path.

        This uses one of the many defined patterns on the options class. But,
        it defaults to a no-op if there are no defined patterns.
        """
        # Iterate through the available patterns.
        for resource, pattern in cls.meta.patterns:
            # Attempt to match the path.
            match = re.match(pattern, path)
            if match is not None:
                # Found something.
                return resource, match.groupdict(), match.string[match.end():]

        # No patterns at all; return unsuccessful.
        return None if not cls.meta.patterns else False

    @classmethod
    def traverse(cls, request, params=None):
        """Traverses down the path and determines the accessed resource.

        This makes use of the patterns array to implement simple traversal.
        This defaults to a no-op if there are no defined patterns.
        """
        # Attempt to parse the path using a pattern.
        result = cls.parse(request.path)
        if result is None:
            # No parsing was requested; no-op.
            return cls, {}

        elif not result:
            # Parsing failed; raise 404.
            raise http.exceptions.NotFound()

        # Partition out the result.
        resource, data, rest = result

        if params:
            # Append params to data.
            data.update(params)

        if resource is None:
            # No traversal; return parameters.
            return cls, data

        # Modify the path appropriately.
        if data.get('path') is not None:
            request.path = data.pop('path')

        elif rest is not None:
            request.path = rest

        # Send us through traversal again.
        result = resource.traverse(request, params=data)
        return result

    @classmethod
    def stream(cls, response, sequence):
        """
        Helper method used in conjunction with the view handler to
        stream responses to the client.
        """
        # Construct the iterator and run the sequence once in order
        # to capture any headers and status codes set.
        iterator = iter(sequence)
        data = {'chunk': next(iterator)}
        response.streaming = True

        def streamer():
            # Iterate through the iterator and yield its content
            while True:
                if response.asynchronous:
                    # Yield our current chunk.
                    yield data['chunk']

                else:
                    # Write the chunk to the response
                    response.send(data['chunk'])

                    # Yield its body
                    yield response.body

                    # Unset the body.
                    response.body = None

                try:
                    # Get the next chunk.
                    data['chunk'] = next(iterator)

                except StopIteration:
                    # Get out of the loop.
                    break

            if not response.asynchronous:
                # Close the response.
                response.close()

        # Return the streaming function.
        return streamer()

    @utils.boundmethod
    def deserialize(self, request=None, text=None, format=None):
        """Deserializes the text using a determined deserializer.

        @param[in] request
            The request object to pull information from; normally used to
            determine the deserialization format (when `format` is
            not provided).

        @param[in] text
            The text to be deserialized. Can be left blank and the
            request will be read.

        @param[in] format
            A specific format to deserialize in; if provided, no detection is
            done. If not provided, the content-type header is looked at to
            determine an appropriate deserializer.

        @returns
            A tuple of the deserialized data and an instance of the
            deserializer used.
        """
        if isinstance(self, Resource):
            if not request:
                # Ensure we have a response object.
                request = self._request

        Deserializer = None
        if format:
            # An explicit format was given; do not attempt to auto-detect
            # a deserializer.
            Deserializer = self.meta.deserializers[format]

        if not Deserializer:
            # Determine an appropriate deserializer to use by
            # introspecting the request object and looking at
            # the `Content-Type` header.
            media_ranges = request.get('Content-Type')
            if media_ranges:
                # Parse the media ranges and determine the deserializer
                # that is the closest match.
                media_types = six.iterkeys(self._deserializer_map)
                media_type = mimeparse.best_match(media_types, media_ranges)
                if media_type:
                    format = self._deserializer_map[media_type]
                    Deserializer = self.meta.deserializers[format]

            else:
                # Client didn't provide a content-type; we're supposed
                # to auto-detect.
                # TODO: Implement this.
                pass

        if Deserializer:
            try:
                # Attempt to deserialize the data using the determined
                # deserializer.
                deserializer = Deserializer()
                data = deserializer.deserialize(request=request, text=text)
                return data, deserializer

            except ValueError:
                # Failed to deserialize the data.
                pass

        # Failed to determine a deserializer; or failed to deserialize.
        raise http.exceptions.UnsupportedMediaType()

    @utils.boundmethod
    def serialize(self, data, response=None, request=None, format=None):
        """Serializes the data using a determined serializer.

        @param[in] data
            The data to be serialized.

        @param[in] response
            The response object to serialize the data to.
            If this method is invoked as an instance method, the response
            object can be omitted and it will be taken from the instance.

        @param[in] request
            The request object to pull information from; normally used to
            determine the serialization format (when `format` is not provided).
            May be used by some serializers as well to pull additional headers.
            If this method is invoked as an instance method, the request
            object can be omitted and it will be taken from the instance.

        @param[in] format
            A specific format to serialize in; if provided, no detection is
            done. If not provided, the accept header (as well as the URL
            extension) is looked at to determine an appropriate serializer.

        @returns
            A tuple of the serialized text and an instance of the
            serializer used.
        """
        if isinstance(self, Resource):
            if not request:
                # Ensure we have a response object.
                request = self._request

        Serializer = None
        if format:
            # An explicit format was given; do not attempt to auto-detect
            # a serializer.
            Serializer = self.meta.serializers[format]

        if not Serializer:
            # Determine an appropriate serializer to use by
            # introspecting the request object and looking at the `Accept`
            # header.
            media_ranges = (request.get('Accept') or '*/*').strip()
            if not media_ranges:
                # Default the media ranges to */*
                media_ranges = '*/*'

            if media_ranges != '*/*':
                # Parse the media ranges and determine the serializer
                # that is the closest match.
                media_types = six.iterkeys(self._serializer_map)
                media_type = mimeparse.best_match(media_types, media_ranges)
                if media_type:
                    format = self._serializer_map[media_type]
                    Serializer = self.meta.serializers[format]

            else:
                # Client indicated no preference; use the default.
                default = self.meta.default_serializer
                Serializer = self.meta.serializers[default]

        if Serializer:
            try:
                # Attempt to serialize the data using the determined
                # serializer.
                serializer = Serializer(request, response)
                return serializer.serialize(data), serializer

            except ValueError:
                # Failed to serialize the data.
                pass

        # Either failed to determine a serializer or failed to serialize
        # the data; construct a list of available and valid encoders.
        available = {}
        for name in self.meta.allowed_serializers:
            Serializer = self.meta.serializers[name]
            instance = Serializer(request, None)
            if instance.can_serialize(data):
                available[name] = Serializer.media_types[0]

        # Raise a Not Acceptable exception.
        raise http.exceptions.NotAcceptable(available)

    @classmethod
    def _process_cross_domain_request(cls, request, response):
        """Facilitate Cross-Origin Requests (CORs).
        """

        # Step 1
        # Check for Origin header.
        origin = request.get('Origin')
        if not origin:
            return

        # Step 2
        # Check if the origin is in the list of allowed origins.
        if not (origin in cls.meta.http_allowed_origins or
                '*' == cls.meta.http_allowed_origins):
            return

        # Step 3
        # Try to parse the Request-Method header if it exists.
        method = request.get('Access-Control-Request-Method')
        if method and method not in cls.meta.http_allowed_methods:
            return

        # Step 4
        # Try to parse the Request-Header header if it exists.
        headers = request.get('Access-Control-Request-Headers', ())
        if headers:
            headers = [h.strip() for h in headers.split(',')]

        # Step 5
        # Check if the headers are allowed on this resource.
        allowed_headers = [h.lower() for h in cls.meta.http_allowed_headers]
        if any(h.lower() not in allowed_headers for h in headers):
            return

        # Step 6
        # Always add the origin.
        response['Access-Control-Allow-Origin'] = origin

        # TODO: Check if we can provide credentials.
        response['Access-Control-Allow-Credentials'] = 'true'

        # Step 7
        # TODO: Optionally add Max-Age header.

        # Step 8
        # Add the allowed methods.
        allowed_methods = ', '.join(cls.meta.http_allowed_methods)
        response['Access-Control-Allow-Methods'] = allowed_methods

        # Step 9
        # Add any allowed headers.
        allowed_headers = ', '.join(cls.meta.http_allowed_headers)
        if allowed_headers:
            response['Access-Control-Allow-Headers'] = allowed_headers

        # Step 10
        # Add any exposed headers.
        exposed_headers = ', '.join(cls.meta.http_exposed_headers)
        if exposed_headers:
            response['Access-Control-Expose-Headers'] = exposed_headers

    def __init__(self, request, response):
        # Store the request and response objects on self.
        self.request = request
        self.response = response

    def dispatch(self, request, response):
        """Entry-point of the dispatch cycle for this resource.

        Performs common work such as authentication, decoding, etc. before
        handing complete control of the result to a function with the
        same name as the request method.
        """
        # Assert authentication and attempt to get a valid user object.
        self.require_authentication(request)

        # Assert accessibiltiy of the resource in question.
        self.require_accessibility(request.user, request.method)

        # Facilitate CORS by applying various headers.
        # This must be done on every request.
        # TODO: Provide cross_domain configuration that turns this off.
        self._process_cross_domain_request(request, response)

        # Route the HTTP/1.1 request to an appropriate method.
        return self.route(request, response)

    def require_authentication(self, request):
        """Ensure we are authenticated."""
        request.user = user = None

        if request.method == 'OPTIONS':
            # Authentication should not be checked on an OPTIONS request.
            return

        for auth in self.meta.authentication:
            user = auth.authenticate(request)
            if user is False:
                # Authentication protocol failed to authenticate;
                # pass the baton.
                continue

            if user is None and not auth.allow_anonymous:
                # Authentication protocol determined the user is
                # unauthenticated.
                auth.unauthenticated()

            # Authentication protocol determined the user is indeed
            # authenticated (or not); Store the user for later reference.
            request.user = user
            return

        if not user and not auth.allow_anonymous:
            # No authenticated user found and protocol doesn't allow
            # anonymous users.
            auth.unauthenticated()

    def require_accessibility(self, user, method):
        """Ensure we are allowed to access this resource."""
        if method == 'OPTIONS':
            # Authorization should not be checked on an OPTIONS request.
            return

        authz = self.meta.authorization
        if not authz.is_accessible(user, method, self):
            # User is not authorized; raise an appropriate message.
            authz.unaccessible()

    def require_http_allowed_method(cls, request):
        """Ensure that we're allowed to use this HTTP method."""
        allowed = cls.meta.http_allowed_methods
        if request.method not in allowed:
            # The specified method is not allowed for the resource
            # identified by the request URI.
            # RFC 2616 § 10.4.6 — 405 Method Not Allowed
            raise http.exceptions.MethodNotAllowed(allowed)

    def route(self, request, response):
        """Processes every request.

        Directs control flow to the appropriate HTTP/1.1 method.
        """
        # Ensure that we're allowed to use this HTTP method.
        self.require_http_allowed_method(request)

        # Retrieve the function corresponding to this HTTP method.
        function = getattr(self, request.method.lower(), None)
        if function is None:
            # Server is not capable of supporting it.
            raise http.exceptions.NotImplemented()

        # Delegate to the determined function to process the request.
        return function(request, response)

    def options(self, request, response):
        """Process an `OPTIONS` request.

        Used to initiate a cross-origin request. All handling specific to
        CORS requests is done on every request however this method also
        returns a list of available methods.
        """
        # Gather a list available HTTP/1.1 methods for this URI.
        response['Allowed'] = ', '.join(self.meta.http_allowed_methods)

        # All CORS handling is done for every HTTP/1.1 method.
        # No more handling is neccesary; set the response to 200 and return.
        response.status = http.client.OK
